#!bin/bash
#
# This file is part of Lithionite
#
# Copyright (C) 2011  Ivaylo Valkov <ivaylo@e-valkov.org>
#
# Lithionite is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lithionite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lithionite.  If not, see <http://www.gnu.org/licenses/>.
#
# Lithionite is a script/player that handles LETio (http://letio.com)
# audio streams for various Internet radio stations.
#
# Lithionite version 0.1
# http://e-valkov.org/lithionite

VERSION=0.1
RELEASE_DATE="2nd October 2011"


# Set to 1 for debugging messages or use -d
debug=0;

# Require rtmppdump for versions of MPlayer < 1.0rc4
req_rtmpdump=1;

# Missing tools. Only RTMPDump should be listed. When other tools are
# missing the script quits.
missing="";

# Set to 1 if mplayer (>=1.0rc4) supports rtmp.
mplayer_rtmp=0;

function usage
{
    echo "Lithionite $VERSION ($RELEASE_DATE), listen audio streams from LETio powered internet radio stations";
    echo "Usage: lithionite [option] <URL|KEY>";
    echo "";
    echo " -v --version display the version of Lithionite";
    echo " -d --debug --verbose print debugging information"
    echo " -h --help print this help";
    echo "";
    echo "URL is the link to the web page of the live stream";
    echo "KEY is the id of the LETio stream in the form gs=[0-9]+";
    echo "";
    echo "Check http://e-valkov.org/lithionite for newer versions and bug-fixes."  

    exit 1;
}


function check_tools
{
    err=0;
    quit_on_errors=1;

    for tool in grep sed cut mplayer mkfifo wget rm mktemp rtmpdump;
    do
	err=$(${tool} 2>&1 | grep $0:);
	if [ "${err}" ];
	then
	    missing=${missing}" ${tool}";
	    err=1;

	    if [ "${tool}" = "rtmpdump" ];
	    then
		# RTMPDump migth not be needed at all. It is not a
		# good idea to exit just if it is missing.
		quit_on_errors=0;
	    fi
	else
	    err=0;
	fi

	if [ "${tool}" = "mplayer" ] && [ "${err}" -ne 1 ];
	then
	    m_ver=$(mplayer | grep \
		-E "MPlayer\s*[0-9]+\.[0-9]+[A-Za-z]*[0-9]+" -o \
		| cut -d" " -f2 );
	    major=$(echo $m_ver| cut -d"." -f1);
	    minor=$(echo $m_ver | grep -E "\.[0-9]+" -o | cut -d"." -f2);
	    rc=$(echo $m_ver | grep -E "rc[0-9]+" -o );
	    rc_digit=$(echo $rc| grep -E "[0-9]+" -o);

	    # Mplayer 1.0.rc4 and above supports rtmp streams
	    if [ "$rc" ] && [ "$rc_digit" -ge 4 ] && \
		[ "$major" -eq 1 ] && [ "$minor" -eq 0 ];
	    then
		if [ "${debug}" -eq 1 ];
		then
		    echo "Mplayer version 1.0 >=rc4" > /dev/stderr;
		fi
		req_rtmpdump=0;
		mplayer_rtmp=1;
	    elif [ "$major" -eq 1 ] && [ "$minor" -gt 0 ];
	    then
		echo "Mplayer version >= 1.0rc4" > /dev/stderr;
		req_rtmpdmp=0;
		mplayer_rtmp=1;
	    fi
	fi

	if [ "${tool}" = "rtmpdump" ] && [ "${req_rtmpdump}" -eq 0 ] && \
	    [ "${err}" ];
	then
	    # Mplayer 1.0.rc4 and above supports rtmp streams
	    if [ "${debug}" -eq 1 ];
	    then
		echo "Removing rtmpdump requirement. MPlayer should support RTMP streams." 2> /dev/stderr;
	    fi
	    err=0;
	fi
    done

    if [ "${err}" -eq 1 ];
    then
	echo "Missing ${missing}" 2> /dev/null;
	if [ "$quit_on_errors" -eq 1 ];
	then
	    exit 3;
	fi
    fi
}

function extract_gs_from_http_response
{
    # $1 - URL
    wget -O - "$1" 2>/dev/null | \
	grep LETio_divLP |grep -E "gs=[0-9]+" -o |cut -d'=' -f2 ;
}

function extract_gs_from_url_params
{
    # $1 - URL
    echo "$1" | grep -E "gs=[0-9]+" -o |cut -d'=' -f2
}

function clean_tmp_files
{
    rm ${meta_data} ${buf1} ${buf2} ${mplayer_input};
    exit 0;
}


function playback_loop
{
    # $1 - gs/station key 

    for gs in $1;
    do 
	meta_url="http://master.letio.com/getMD.aspx?gs=${gs}"; 
	stream=${gs}; 

	station="";
	station_url="";
	track="";
	was_playing="";
	now_playing="";

	buf_count=1;

	mplayer_pid=-1;

	while [ -n "${stream}" ] ; 
	do 
	    # Keeps the previous value of $stream.
	    first_run="${stream}";

	    if [ "${debug}" -eq 1 ];
	    then
		echo "Fetching metadata to ${meta_data}" > /dev/stderr;
	    fi

	    # Fetch metadata
	    wget -O ${meta_data} "${meta_url}" 2>/dev/null;

	    # Extract audio URL
	    stream=$(grep urlMusicPart ${meta_data} \
		| sed -e 's/urlMusicPart=//' \
		-e "s/\s$//"  \
		-e 's/\s/%20/' \
		-e 's/\x0D//g');
	             # ^^^ Remove weird invalid hidden character.

	    station=$(grep station= ${meta_data}| cut -d"=" -f2);
	    station_url=$(grep url= ${meta_data} | cut -d"=" -f2);
	    # It is unclear which record is the current song if there
	    # are several records. The last one is taken blindly.
	    track=$(grep mdd= ${meta_data} | cut -d"|" -f1,3| \
		sed -e "s/mdd=[0-9]\+-[0-9]\+-[0-9]\+\s*//g" | tail -n1| cut -d"|" -f2);

	    if [ "$station" ];
	    then
		now_playing="Station: ${station}";
	    fi

	    if [ "${station_url}" ];
	    then
		now_playing=${now_playing}"\nURL: ${station_url}";
	    fi

	    if [ "${track}" ];
	    then
		now_playing=${now_playing}"\nNow playing (probably?): ${track}";
	    fi

	    if [ "${now_playing}" != "${was_playing}" ];
	    then
		echo -e "${now_playing}";
		was_playing=${now_playing}
	    fi

	    if [ "${debug}" -eq 1 ];
	    then
		echo "Audio stream URL for next request: ${stream}"
	    fi

	    # Timestamp/key for the next part of the stream.
	    elp=$(grep guidMetaNext ${meta_data}| cut -d'=' -f2);
	    meta_url=$(echo "${meta_url}" \
		| sed -e 's/&elp.*//g')"&elp=${elp}";

	    if [ "${debug}" -eq 1 ];
	    then
		echo "Metadata URL for next request: ${meta_url}" > /dev/stderr;
	    fi


	    if [ "${buf_count}"  -ge 3 ]; 
	    then
		buf_count=1;
	    fi

	    if [ "${buf_count}" -eq 1 ];
	    then
		buffer=${buf1};
	    elif [ "${buf_count}" -eq 2 ];
	    then
		buffer=${buf2};
	    fi

	    rtmp_stream=$(echo "$stream" | grep -E "^rtmp:");

	    if [ "$rtmp_stream" ];
	    then
		if [ "$req_rtmpdump" -eq 1 ];
		then
		    # RTMPDump 

		    if [ "${debug}" -eq 1 ];
		    then
			echo "Using RTMPDump for RTMP stream" 2> /dev/stderr
		    fi
		    err=$(echo $missing | grep rtmpdump);
		    if [ "${err}" ];
		    then
			echo "RTMPDump is missing. The RTMP stream could not be handled." > /dev/stderr;
 			# Quit
			break;
		    fi

		    # Don't wait for the download, it will take more
		    # than 10 minutes
		    rtmpdump -o ${buffer} -r ${stream} 2> /dev/null >/dev/null &
		    mplayer_parameters=" ${buffer} -nocache -novideo";

		else
		    # Mplayer alone
		    if [ "${debug}" -eq 1 ];
		    then
			echo "Using Mplayer (>=1.0rc4) for RTMP stream" 2> /dev/stderr
		    fi
		    
		    if [ "${mplayer_rtmp}" -eq 0 ];
		    then
			echo "It seems this version of mplayer does not support RTMP streams. At least version 1.0rc4 is needed." > /dev/stderr;
			# Quit
			break;
		    fi
		    # Cacheing is slow. Video seariching is slow and
		    # it is not present.
		    mplayer_parameters=" -nocache -novideo ${stream}";
		fi
	    else


		if [ "${debug}" -eq 1 ];
		then
		    echo "Fetching next 44 seconds to ${buffer}" > /dev/stderr;
		fi

		wget -O "${buffer}" "${stream}" 2> /dev/null ;

		if [ "${first_run}" != "${gs}" ];
		then
		    echo "loadfile ${buffer}" \
			> ${mplayer_input} &
		fi
	    fi


	    # Run Mplayer only once
	    if [ "$mplayer_pid" -eq -1 ];
	    then
		if [ "${debug}" -eq 1 ];
		then
		    echo "Starting mplayer" > /dev/stderr;
		fi

		if [ "$req_rtmpdump" -eq 1 ] || [ ! "${mplayer_parameters}" ];
		then
		    echo "Waiting 5 seconds for buffers to fill up.";
		    sleep 5;
		fi

		if [ "$mplayer_parameters" ] ;
		then
		    echo "Starting mplayer with rtmp stream. This may take a while" > /dev/stderr;
		    mplayer -slave -quiet ${mplayer_parameters} \
			-input file=${mplayer_input} 2> /dev/null \
			> /dev/null </dev/null &
		else
		    mplayer -slave -quiet ${buffer} \
			-input file=${mplayer_input} 2> /dev/null \
			> /dev/null </dev/null &
		fi
		mplayer_pid=$!;
	    fi

	    # RTMP streams do not need more work just sleep infinitely
	    if [ "$rtmp_stream" ];
	    then
		while [ 1 ] ;
		do
		    sleep 60;
		done
	    fi
	    
	    let buf_count++;
	    # Next request must be done after ~44 sec. 42 works
	    # too. With the five second delay before starting mplayer,
	    # 37 is OK.
	    sleep 37;
	done ;
    done
}


if [ ! "$1" ] 
then
    usage;
fi

if [ "$2" = "-d" ] || [ "$2" = "--debug" ] || [ "$2" = "--verbose" ] || \
    [ "$1" = "-d" ] || [ "$1" = "--debug" ] || [ "$1" = "--verbose" ];
then
    debug=1;
fi

if [ "$2" = "-h" ] || [ "$2" = "--help" ]  || \
    [ "$1" = "-h" ] || [ "$1" = "--help" ];
then
    usage;
fi

if [ "$2" = "-v" ] || [ "$2" = "--version" ] || \
    [ "$1" = "-v" ] || [ "$1" = "--version" ];
then
    echo -e "Lithionite version $VERSION ($RELEASE_DATE)\n";
    # license
    echo -e $(head -n 19 $0 | tail -n 15 | sed -e 's/^#\s*/\\n/g');
    exit 0;
fi

if [ ! "$(echo $1|grep ^http://)" ]  && [ ! "$(echo $2 | grep ^http://)" ] && \
    [ ! "$(echo $1| grep -E 'gs=[0-9]+')" ] && [ ! "$(echo $2| grep -E 'gs=[0-9]+')" ];
then
    echo "Not a valid URL or KEY";
    exit 2;
elif [ "$(echo $1|grep ^http://)" ] || [ "$(echo $1 | grep -E 'gs=[0-9]+')" ] ;
then
    URL="$1";
else
    URL="$2";
fi

check_tools;

if [ "${debug}" -eq 1 ];
then
    echo -n "Trying to extract key from URL params... " > /dev/stderr;
fi

gs=$(extract_gs_from_url_params "${URL}");

if [ ! "$gs" ];
then
    if [ "${debug}" -eq 1 ];
    then
	echo -ne "\nTrying to extract key from HTML... " > /dev/stderr;
    fi

    gs=$(extract_gs_from_http_response "${URL}");
fi

if [ "$gs" ];
then

    if [ "${debug}" -eq 1 ];
    then
	echo "SUCCESS";
    fi

    meta_data=$(mktemp -t lithionite_meta_data_XXX.tmp);

    # Control MPlayer over FIFO.
    mplayer_input=$(mktemp -t lithionite_mplayer_input_XXX.tmp);
    rm ${mplayer_input};
    mkfifo ${mplayer_input};
    
    # Buf1 
    buf1=$(mktemp -t lithionite_buf1_XXX.tmp);
    # Buf2 
    buf2=$(mktemp -t lithionite_buf2_XXX.tmp);

    # Capture signals
    trap 'clean_tmp_files' SIGKILL SIGTERM SIGINT
    
    playback_loop "$gs"
    # Normal clean
    clean_tmp_files;
else
    echo "No key found" > /dev/stderr;
    exit 4;
fi
